// ModemSim clas interface: run a simulation
//
// CONSTRUCTION: with three parameters: the number of
//     modems, the average connect time, and the
//     interarrival time
//
// ******************PUBLIC OPERATIONS*********************
// void runSim( )       --> Run a simulation

import DataStructures.*;
import Supporting.*;
import Exceptions.*;
import Supporting.Comparable;

/**
 * The event class.
 * Implements the Comparable interface
 * to arrange events by time of occurrence.
 */
class Event implements Comparable
{
    static final int DIAL_IN = 1;
    static final int HANGUP  = 2;

    Event( )
    {
        this( 0, 0, DIAL_IN );
    }

    Event( int name, long tm, int type )
    {
        who  = name;
        time = tm;
        what = type;
    }


    public boolean lessThan( Comparable rhs )
      { return time < ( (Event) rhs ).time; }

    public int compares( Comparable rhs )
    {
        return lessThan( rhs ) ? -1 : rhs.lessThan( this ) ? 1 : 0;
    }
  
    int who;        // the number of the user
    long time;      // when the event will occur
    int what;       // DIAL_IN or HANGUP
}

/**
 * The ModemSim class.
 */
public class ModemSim
{
    /**
     * Constructor.
     * @param modem number of modems.
     * @param avgLen averge length of a call.
     * @param callIntrvl the average time between calls.
     */
    public ModemSim( int modems, double avgLen, long callIntrvl )
    {
        eventSet    = new BinaryHeap( new Event( ) );
        freeModems  = modems;
        avgCallLen  = avgLen;
        freqOfCalls = callIntrvl;
        r           = new Random( );
        nextCall( freqOfCalls );  // Schedule first call
    }

    private Random r;                 // A random source
    private PriorityQueue eventSet;   // Pending events

        // Basic parameters of the simulation
    private int freeModems;           // Number of modems unused
    private double avgCallLen;        // Length of a call
    private long freqOfCalls;         // Interval between calls

        // Used by nextCall only
    private int  userNum = 0;
    private long nextCallTime = 0;

    /**
     * Place a new DIAL_IN event into the event queue.
     * Then advance the time when next DIAL_IN event will occur.
     * In practice, we would use a random number to set the time.
     */
    private void nextCall( long delta )
    {
        eventSet.insert( new Event( userNum++, nextCallTime, Event.DIAL_IN ) );
        nextCallTime += delta;
    }


    /**
     * Run the simulation until stoppingTime occurs.
     * Print output as in Figure 4.
     */
    public void runSim( long stoppingTime )
    {
        Event e = null;
        long howLong;

        while( !eventSet.isEmpty( ) )
        {
            try
              { e = (Event) eventSet.deleteMin( );
            }
            catch( Underflow ex ) { } // Cannot happen

            if( e.time > stoppingTime )
                break;

            if( e.what == Event.HANGUP )    // HANGUP
            {
                freeModems++;
                System.out.println( "User " + e.who + " hangs up at time " + e.time );
            }
            else                      // DIAL_IN
            {
                System.out.print(  "User " + e.who +" dials in at time "
                                 + e.time + " " );
                if( freeModems > 0 )
                {
                    freeModems--;
                    howLong = r.poisson( avgCallLen );
                    System.out.println(  "and connects for " + howLong + " minutes" );
                    e.time += howLong;
                    e.what = Event.HANGUP;
                    eventSet.insert( e );
                }
                else
                    System.out.println( "but gets busy signal" );

                nextCall( freqOfCalls );
            }
        }
    }

    /**
     * Quickie main for testing purposes.
     */
    public static void main( String [ ] args )
    {
        ModemSim s = new ModemSim( 3, 5.0, 1 );
        s.runSim( 20 );
    }
}
